#ifdef LAME_MT

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>

#include "lame.h"
#include "util.h"
#include "newmdct.h"
#include "psymodel.h"
#include "quantize.h"
#include "quantize_pvt.h"
#include "bitstream.h"
#include "VbrTag.h"
#include "vbrquantize.h"

#include "lame_global_flags.h"
#include "lame_mt.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

#define USE_THREADS

int is_smp_system = 0;
MT_THREAD_TYPE param_phase_1;
int stub[128];
MT_THREAD_TYPE param_phase_2;

#if defined(_WIN32) && defined(USE_THREADS)
HANDLE sp1,sp2,ep[2];
HANDLE hthread[2];
CRITICAL_SECTION param_mutex;



DWORD WINAPI encode_phase_1 (LPVOID t) {

	while(1)
	{
		MT_THREAD_TYPE *x=&param_phase_1;
		WaitForSingleObject(sp1,INFINITE);
		if (x->kill)
			return 0;
		lame_encode_mp3_frame_phase_1(x->gfp,x->inbuf_l,x->inbuf_r,x->mp3buf,x->mp3buf_size,x->gbuff);
		SetEvent(ep[0]);
	}
}

DWORD WINAPI encode_phase_2 (LPVOID t) {
	
	while(1)
	{
		MT_THREAD_TYPE *x=&param_phase_2;
		WaitForSingleObject(sp2,INFINITE);
		if (x->kill)
			return 0;
		x->ret = lame_encode_mp3_frame_phase_2(x->gfp, x->inbuf_l, x->inbuf_r, x->mp3buf, x->mp3buf_size,x->gbuff);
		SetEvent(ep[1]);

	}
}

#endif

void adjust_ATH( lame_global_flags* const  gfp,
            FLOAT              tot_ener[2][4] );

void updateStats( lame_internal_flags * const gfc );

FLOAT pecalc_s( III_psy_ratio *mr, FLOAT masking_lower );
FLOAT pecalc_l( III_psy_ratio *mr, FLOAT masking_lower );

int Compute_Value_Of_PE( lame_global_flags * gfp,int gr_out, 
		    III_psy_ratio masking_ratio[2][2],
		    III_psy_ratio masking_MS_ratio[2][2],
		    FLOAT percep_entropy[2],FLOAT percep_MS_entropy[2], 
                    int blocktype_d[2])
{
/* to get a good cache performance, one has to think about
 * the sequence, in which the variables are used.  
 * (Note: these static variables have been moved to the gfc-> struct,
 * and their order in memory is layed out in util.h)
 */
    lame_internal_flags *gfc=gfp->internal_flags;

    /* usual variables like loop indices, etc..    */
    int numchn, chn;

    numchn = gfc->channels_out;
    /* chn=2 and 3 = Mid and Side channels */
    if (gfp->mode == JOINT_STEREO) numchn=4;

    /*********************************************************************
     * compute the value of PE to return ... no delay and advance
     *********************************************************************/
    for(chn=0;chn<numchn;chn++) {
	FLOAT *ppe;
	int type;
	III_psy_ratio *mr;

	if (chn > 1) {
	    ppe = percep_MS_entropy - 2;
	    type = NORM_TYPE;
	    if (blocktype_d[0] == SHORT_TYPE || blocktype_d[1] == SHORT_TYPE)
		type = SHORT_TYPE;
	    mr = &masking_MS_ratio[gr_out][chn-2];
	} else {
	    ppe = percep_entropy;
	    type = blocktype_d[chn];
	    mr = &masking_ratio[gr_out][chn];
	}

	if (type == SHORT_TYPE)
	    ppe[chn] = pecalc_s(mr, gfc->masking_lower);
	else
	    ppe[chn] = pecalc_l(mr, gfc->masking_lower);

#if defined(HAVE_GTK)
	if (gfp->analysis) gfc->pinfo->pe[gr_out][chn] = ppe[chn];
#endif
    }
    return 0;
}

int lame_encode_mp3_frame_phase_1 (		// Output
	lame_global_flags* const  gfp,			// Context
	sample_t*                 inbuf_l,      // Input
	sample_t*                 inbuf_r,      // Input
	unsigned char*            mp3buf, 		// Output
	int                    mp3buf_size,		// Output
	LAME_MT_BUFFER_TYPE*	 goutp	)		// Which buffer to use
{
  const sample_t *inbuf[2];
  lame_internal_flags *gfc=gfp->internal_flags;
  int ch,gr;

  inbuf[0]=inbuf_l;
  inbuf[1]=inbuf_r;

  if (gfc->lame_encode_frame_init==0 ) {
      /* prime the MDCT/polyphase filterbank with a short block */
      int i,j;
      sample_t primebuff0[286+1152+576];
      sample_t primebuff1[286+1152+576];
      gfc->lame_encode_frame_init=1;
      for (i=0, j=0; i<286+576*(1+gfc->mode_gr); ++i) {
	  if (i<576*gfc->mode_gr) {
	      primebuff0[i]=0;
	      if (gfc->channels_out==2) 
		  primebuff1[i]=0;
	  }else{
	      primebuff0[i]=inbuf[0][j];
	      if (gfc->channels_out==2) 
		  primebuff1[i]=inbuf[1][j];
	      ++j;
	  }
      }
      /* polyphase filtering / mdct */
      for ( gr = 0; gr < gfc->mode_gr; gr++ ) {
	  for ( ch = 0; ch < gfc->channels_out; ch++ ) {
	      goutp->l3_side.tt[gr][ch].block_type=SHORT_TYPE;
	  }
      }
      mdct_sub48(gfc, primebuff0, primebuff1);

      /* check FFT will not use a negative starting offset */
#if 576 < FFTOFFSET
# error FFTOFFSET greater than 576: FFT uses a negative offset
#endif
      /* check if we have enough data for FFT */
      assert(gfc->mf_size>=(BLKSIZE+gfp->framesize-FFTOFFSET));
      /* check if we have enough data for polyphase filterbank */
      assert(gfc->mf_size>=(512+gfp->framesize-32));
  }

  if (gfc->psymodel) {
    /* psychoacoustic model
     * psy model has a 1 granule (576) delay that we must compensate for
     * (mt 6/99).
     */
    int ret;
    const sample_t *bufp[2];  /* address of beginning of left & right granule */
    int blocktype[2];

	// instead of ms_ratio_prev=gfc->ms_ratio[gfc->mode_gr-1];
	// doing special magic
	static double prev;
	static int is_first_iteration = 1;

	if (is_first_iteration) {
		is_first_iteration = 0;
		prev=gfc->ms_ratio[gfc->mode_gr-1];
	}
		
    goutp->ms_ratio_prev=prev;
	// till here - there is more down
    
    for (gr=0; gr < gfc->mode_gr ; gr++) {

    for ( ch = 0; ch < gfc->channels_out; ch++ )
        bufp[ch] = &inbuf[ch][576 + gr*576-FFTOFFSET];

    if (gfp->psymodel == PSY_NSPSYTUNE) {
        ret=L3psycho_anal_ns( gfp, bufp, gr, 
                            &goutp->ms_ratio[gr],&goutp->ms_ratio_next,
                            goutp->masking_LR, goutp->masking_MS,
                            goutp->pe[gr],goutp->pe_MS[gr],goutp->tot_ener[gr],blocktype);
    } else {
        ret=L3psycho_anal( gfp, bufp, gr, 
                            &goutp->ms_ratio[gr],&goutp->ms_ratio_next,
                            goutp->masking_LR, goutp->masking_MS,
                            goutp->pe[gr],goutp->pe_MS[gr],goutp->tot_ener[gr],blocktype);
    }
	
    if (ret!=0) {
		return -4;
	}

	// instead of ms_ratio_prev=gfc->ms_ratio[gfc->mode_gr-1];
	// continue special !!!
	prev = goutp->ms_ratio[gfc->mode_gr-1];
	// till here

      if (gfp->mode == JOINT_STEREO) {
	  goutp->ms_ener_ratio[gr] = goutp->tot_ener[gr][2]+goutp->tot_ener[gr][3];
	  if (goutp->ms_ener_ratio[gr]>0)
	      goutp->ms_ener_ratio[gr] = goutp->tot_ener[gr][3]/goutp->ms_ener_ratio[gr];
      }

      /* block type flags */
      for ( ch = 0; ch < gfc->channels_out; ch++ ) {
	  // instead of this::
	  /*gr_info *cod_info = &gfc->l3_side.tt[gr][ch];
	  cod_info->block_type=blocktype[ch];
	  cod_info->mixed_block_flag = 0;*/
	  goutp->l3_side.tt[gr][ch].block_type=blocktype[ch];
	  goutp->l3_side.tt[gr][ch].mixed_block_flag=0;
	  // till here
      }
    }
  }

  return 0;
}

int  lame_encode_mp3_frame_phase_2 (				// Output
	lame_global_flags* const  gfp,			// Context
	sample_t*                 inbuf_l,              // Input
	sample_t*                 inbuf_r,              // Input
	unsigned char*            mp3buf, 		// Output
	int                    mp3buf_size,		// Output
	LAME_MT_BUFFER_TYPE*	 ginp	)		// Which buffer to use
{
  int mp3count;
  III_psy_ratio masking_LR[2][2];    /*LR masking & energy */
  III_psy_ratio masking_MS[2][2]; /*MS masking & energy */
  III_psy_ratio (*masking)[2][2];  /*pointer to selected maskings*/
  const sample_t *inbuf[2];
  lame_internal_flags *gfc=gfp->internal_flags;

  FLOAT tot_ener[2][4];
  FLOAT ms_ener_ratio[2]={.5,.5};
  chgrdata pe,pe_MS;
  chgrdata *pe_use;

  int ch,gr;

  FLOAT ms_ratio_next = 0.;
  FLOAT ms_ratio_prev = 0.;

  // Computing the value of PE as should have been done in phase I
  if (gfc->psymodel) {
    int blocktype[2];
    for (gr=0; gr < gfc->mode_gr ; gr++) {
		if (gfp->psymodel == PSY_NSPSYTUNE) {        

			for ( ch = 0; ch < gfc->channels_out; ch++ )
				blocktype[ch] = ginp->l3_side.tt[gr][ch].block_type;

			Compute_Value_Of_PE( gfp, gr,ginp->masking_LR, ginp->masking_MS,
                            ginp->pe[gr],ginp->pe_MS[gr],blocktype);
		}
	}
  }

  // Read from buffer 'ginp' the output of lame_encode_mp3_frame_phase_1

  memcpy((char *) masking_LR,		(char *) ginp->masking_LR,		sizeof(masking_LR));
  memcpy((char *) masking_MS,		(char *) ginp->masking_MS,		sizeof(masking_MS));

  for (gr=0; gr < gfc->mode_gr ; gr++) {
	  for ( ch = 0; ch < gfc->channels_out; ch++ ) {
		gfc->l3_side.tt[gr][ch].block_type = ginp->l3_side.tt[gr][ch].block_type;
		gfc->l3_side.tt[gr][ch].mixed_block_flag = ginp->l3_side.tt[gr][ch].mixed_block_flag;
	  }
  }

  ms_ratio_next = ginp->ms_ratio_next;
  ms_ratio_prev = ginp->ms_ratio_prev;

  memcpy((char *) ms_ener_ratio,	(char *) ginp->ms_ener_ratio,	sizeof(ms_ener_ratio));
  memcpy((char *) &gfc->ms_ratio,	(char *) &ginp->ms_ratio, 		sizeof(gfc->ms_ratio));
  memcpy((char *) pe,				(char *) ginp->pe,				sizeof(pe));
  memcpy((char *) pe_MS,			(char *) ginp->pe_MS,			sizeof(pe_MS));
  memcpy((char *) tot_ener,			(char *) ginp->tot_ener	,		sizeof(tot_ener));  

  // Done reading the output of lame_encode_mp3_frame_phase_1

  inbuf[0]=inbuf_l;
  inbuf[1]=inbuf_r;

  /********************** padding *****************************/
  /* padding method as described in 
   * "MPEG-Layer3 / Bitstream Syntax and Decoding"
   * by Martin Sieler, Ralph Sperschneider
   *
   * note: there is no padding for the very first frame
   *
   * Robert Hegemann 2000-06-22
   */
  gfc->padding = FALSE;
  if ((gfc->slot_lag -= gfc->frac_SpF) < 0) {
      gfc->slot_lag += gfp->out_samplerate;
      gfc->padding = TRUE;
  }

  if (gfc->psymodel) {
    /*int blocktype[2];
    for (gr=0; gr < gfc->mode_gr ; gr++) {
		if (gfp->psymodel == PSY_NSPSYTUNE) {        

			for ( ch = 0; ch < gfc->channels_out; ch++ )
				blocktype[ch] = gfc->l3_side.tt[gr][ch].block_type;

			Compute_Value_Of_PE( gfp, gr,masking_LR,masking_MS,
				                 pe[gr],pe_MS[gr],blocktype);
		}
	}*/
  } else {
    memset((char *) masking_LR, 0, sizeof(masking_LR));
    memset((char *) masking_MS, 0, sizeof(masking_MS));
    for (gr=0; gr < gfc->mode_gr ; gr++)
      for ( ch = 0; ch < gfc->channels_out; ch++ ) {
	gfc->l3_side.tt[gr][ch].block_type=NORM_TYPE;
	gfc->l3_side.tt[gr][ch].mixed_block_flag=0;
	pe_MS[gr][ch]=pe[gr][ch]=700;
      }
  }

  /* auto-adjust of ATH, useful for low volume */
  adjust_ATH( gfp, tot_ener );



  /* polyphase filtering / mdct */
  mdct_sub48(gfc, inbuf[0], inbuf[1]);

  /* Here will be selected MS or LR coding of the 2 stereo channels */
  gfc->mode_ext = MPG_MD_LR_LR;
  
  if (gfp->force_ms) {
    gfc->mode_ext = MPG_MD_MS_LR;
  } else if (gfp->mode == JOINT_STEREO) {
    int check_ms_stereo = 1;
    /* ms_ratio = is scaled, for historical reasons, to look like
       a ratio of side_channel / total.
       0 = signal is 100% mono
       .5 = L & R uncorrelated
    */

    /* [0] and [1] are the results for the two granules in MPEG-1,
     * in MPEG-2 it's only a faked averaging of the same value
     * _prev is the value of the last granule of the previous frame
     * _next is the value of the first granule of the next frame
     */
    if (gfp->psymodel == PSY_GPSYCHO) {
      FLOAT8  ms_ratio_ave1;
      FLOAT8  ms_ratio_ave2;
      FLOAT8  threshold1    = 0.35;
      FLOAT8  threshold2    = 0.45;

      /* take an average */
      if (gfc->mode_gr==1) {
	  /* MPEG2 - no second granule */
	  ms_ratio_ave1 = 0.33 * ( gfc->ms_ratio[0] + ms_ratio_prev + ms_ratio_next );
	  ms_ratio_ave2 = gfc->ms_ratio[0];
      }else{
	  ms_ratio_ave1 = 0.25 * ( gfc->ms_ratio[0] + gfc->ms_ratio[1] + ms_ratio_prev + ms_ratio_next );
	  ms_ratio_ave2 = 0.50 * ( gfc->ms_ratio[0] + gfc->ms_ratio[1] );
      }
      

      if (ms_ratio_ave1 >= threshold1 || ms_ratio_ave2 >= threshold2)
	check_ms_stereo = 0;
    }
    if (check_ms_stereo) {
      FLOAT8 sum_pe_MS = 0;
      FLOAT8 sum_pe_LR = 0;
      for ( gr = 0; gr < gfc->mode_gr; gr++ ) {
	for ( ch = 0; ch < gfc->channels_out; ch++ ) {
	  sum_pe_MS += pe_MS[gr][ch];
	  sum_pe_LR += pe[gr][ch];
	}
      }
/*2DO rh 20021015 
change the following to
      if (sum_pe_MS <= sum_pe_LR)
	gfc->mode_ext = MPG_MD_MS_LR;
*/
      /* based on PE: M/S coding would not use much more bits than L/R */
      if (((gfp->psymodel == PSY_GPSYCHO) && sum_pe_MS <= 1.07 * sum_pe_LR)
	  || ((gfp->psymodel == PSY_NSPSYTUNE) && sum_pe_MS <= 1.00 * sum_pe_LR)) {
	      gr_info *gi0 = &gfc->l3_side.tt[0][0];
	      gr_info *gi1 = &gfc->l3_side.tt[gfc->mode_gr-1][0];
	      
          if (gi0[0].block_type == gi0[1].block_type
	          && gi1[0].block_type == gi1[1].block_type)
	          gfc->mode_ext = MPG_MD_MS_LR;
      }
    }
  }

  /* bit and noise allocation */
  if (gfc->mode_ext == MPG_MD_MS_LR) {
      masking = &masking_MS;    /* use MS masking */
      pe_use = &pe_MS;
  } else {
      masking = &masking_LR;    /* use LR masking */
      pe_use = &pe;
  }


#if defined(HAVE_GTK)
  /* copy data for MP3 frame analyzer */
  if (gfp->analysis && gfc->pinfo != NULL) {
    for ( gr = 0; gr < gfc->mode_gr; gr++ ) {
      for ( ch = 0; ch < gfc->channels_out; ch++ ) {
	gfc->pinfo->ms_ratio[gr]=gfc->ms_ratio[gr];
	gfc->pinfo->ms_ener_ratio[gr]=ms_ener_ratio[gr];
	gfc->pinfo->blocktype[gr][ch]=gfc->l3_side.tt[gr][ch].block_type;
	gfc->pinfo->pe[gr][ch]=(*pe_use)[gr][ch];
	memcpy(gfc->pinfo->xr[gr][ch], &gfc->l3_side.tt[gr][ch].xr,
	       sizeof(FLOAT8)*576);
	/* in psymodel, LR and MS data was stored in pinfo.  
	   switch to MS data: */
	if (gfc->mode_ext==MPG_MD_MS_LR) {
	  gfc->pinfo->ers[gr][ch]=gfc->pinfo->ers[gr][ch+2];
	  memcpy(gfc->pinfo->energy[gr][ch],gfc->pinfo->energy[gr][ch+2],
		 sizeof(gfc->pinfo->energy[gr][ch]));
	}
      }
    }
  }
#endif

  if (gfp->psymodel == PSY_NSPSYTUNE) {
      if (gfp->VBR == vbr_off || gfp->VBR == vbr_abr) {
        static FLOAT fircoef[9] = {
	    -0.0207887 *5,	-0.0378413*5,	-0.0432472*5,	-0.031183*5,
	    7.79609e-18*5,	 0.0467745*5,	 0.10091*5,	0.151365*5,
	    0.187098*5
        };

        int i;
        FLOAT f;

        for(i=0;i<18;i++) gfc->nsPsy.pefirbuf[i] = gfc->nsPsy.pefirbuf[i+1];

        f = 0.0;
        for ( gr = 0; gr < gfc->mode_gr; gr++ )
          for ( ch = 0; ch < gfc->channels_out; ch++ )
	      f += (*pe_use)[gr][ch];
        gfc->nsPsy.pefirbuf[18] = f;

        f = gfc->nsPsy.pefirbuf[9];
        for (i=0;i<9;i++)
	    f += (gfc->nsPsy.pefirbuf[i]+gfc->nsPsy.pefirbuf[18-i]) * fircoef[i];

        f = (670*5*gfc->mode_gr*gfc->channels_out)/f;
        for ( gr = 0; gr < gfc->mode_gr; gr++ ) {
          for ( ch = 0; ch < gfc->channels_out; ch++ ) {
	    (*pe_use)[gr][ch] *= f;
          }
        }
      }
  }

  switch (gfp->VBR){ 
  default:
  case vbr_off:
    iteration_loop( gfp,*pe_use,ms_ener_ratio, *masking);
    break;
  case vbr_mt:
  case vbr_rh:
  case vbr_mtrh:
    VBR_iteration_loop( gfp,*pe_use,ms_ener_ratio, *masking);
    break;
  case vbr_abr:
    ABR_iteration_loop( gfp,*pe_use,ms_ener_ratio, *masking);
    break;
  }

  /*  write the frame to the bitstream  */
  format_bitstream(gfp);

  /* copy mp3 bit buffer into array */
  mp3count = copy_buffer(gfc,mp3buf,mp3buf_size,1);




  if (gfp->bWriteVbrTag) AddVbrFrame(gfp);


#if defined(HAVE_GTK)
  if (gfp->analysis && gfc->pinfo != NULL) {
    for ( ch = 0; ch < gfc->channels_out; ch++ ) {
      int j;
      for ( j = 0; j < FFTOFFSET; j++ )
	gfc->pinfo->pcmdata[ch][j] = gfc->pinfo->pcmdata[ch][j+gfp->framesize];
      for ( j = FFTOFFSET; j < 1600; j++ ) {
	gfc->pinfo->pcmdata[ch][j] = inbuf[ch][j-FFTOFFSET];
      }
    }
    set_frame_pinfo (gfp, *masking);
  }
#endif
  
#ifdef BRHIST
  updateStats( gfc );
#endif

  return mp3count;
}



void set_mt_mode(lame_global_flags *gfp, mt_modes mode)
{
	if (mode != MT_DEFAULT)
		gfp->use_mt_encode = (int) mode;
	else
		gfp->use_mt_encode = is_smp_system;
}


/* creates sync objects and start threads */
void start_threading(lame_global_flags *gfp)
{

#if defined(_WIN32) && defined(USE_THREADS)
	DWORD temp;
	
	InitializeCriticalSection(&param_mutex);

	memset(&param_phase_1,0,sizeof(param_phase_1));
	memset(&param_phase_2,0,sizeof(param_phase_2));

	
	sp1 = CreateEvent(NULL,FALSE,FALSE,NULL);
	sp2 = CreateEvent(NULL,FALSE,FALSE,NULL);
	ep[0] = CreateEvent(NULL,FALSE,FALSE,NULL);
	ep[1] = CreateEvent(NULL,FALSE,FALSE,NULL);

	hthread[0] = CreateThread(NULL,0,encode_phase_1,NULL,0,&temp);
	hthread[1] = CreateThread(NULL,0,encode_phase_2,NULL,0,&temp);

	if (!sp1 || !ep[0] || !ep[1] || !sp2 || !hthread[0] || !hthread[1])
		gfp->use_mt_encode = 0;


#endif

}

void stop_theading(lame_global_flags *gfp)
{
#if defined(_WIN32) && defined(USE_THREADS)
	DWORD kill_time;

	param_phase_1.kill = 1;
	param_phase_2.kill = 1;
	SetEvent(sp1);
	SetEvent(sp2);

	kill_time = WaitForMultipleObjects(2,hthread,TRUE,1000);
	
	if (kill_time == WAIT_TIMEOUT)
		fprintf(stderr,"Thread closing issue\n");
	
	CloseHandle(sp1);
	CloseHandle(sp2);
	CloseHandle(ep[0]);
	CloseHandle(ep[1]);

	DeleteCriticalSection(&param_mutex);
#endif
}

int execute_mt_stage(MT_THREAD_TYPE *phase_1_params,MT_THREAD_TYPE *phase_2_params)
{
//#if defined(_WIN32) && defined(USE_THREADS)
#if 1
	/* set parmas (atomic) */

#if 0
	EnterCriticalSection(&param_mutex);
	
	param_phase_1 = *phase_1_params;
	param_phase_2 = *phase_2_params;
	param_phase_1.kill = 0;
	param_phase_2.kill = 0;
	
	LeaveCriticalSection(&param_mutex);

	SetEvent(sp1);
	SetEvent(sp2);

	WaitForMultipleObjects(2,ep,TRUE,INFINITE);
	return param_phase_2.ret;
#else
	int ret;

	param_phase_1 = *phase_1_params;
	param_phase_1.kill = 0;

	SetEvent(sp1);
	
	ret =  lame_encode_mp3_frame_phase_2(phase_2_params->gfp, 
		phase_2_params->inbuf_l, phase_2_params->inbuf_r, 
		phase_2_params->mp3buf, phase_2_params->mp3buf_size,
		phase_2_params->gbuff);

	WaitForSingleObject(ep[0],INFINITE);

	return ret;
#endif

#else
	MT_THREAD_TYPE *x;
	int ret;

	x = phase_2_params;
	ret = lame_encode_mp3_frame_phase_2(x->gfp, x->inbuf_l, x->inbuf_r, x->mp3buf, x->mp3buf_size,x->gbuff);	
	x = phase_1_params;
	lame_encode_mp3_frame_phase_1(x->gfp,x->inbuf_l,x->inbuf_r,x->mp3buf,x->mp3buf_size,x->gbuff);
	
	

	return ret;	
#endif
}

#endif